/* -*- Mode: javascript; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Calendar code.
 *
 * The Initial Developer of the Original Code is
 *   Joey Minta <jminta@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Matthew Willis <mattwillis@gmail.com>
 *   Ferdinand Grassmann <ferdinand@grassmann.info>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Prints a rough month-grid of events/tasks
 */

function FGcalYearPrinter() {
    this.wrappedJSObject = this;
}

FGcalYearPrinter.prototype.QueryInterface =
function QueryInterface(aIID) {
    if (!aIID.equals(Components.interfaces.nsISupports) &&
        !aIID.equals(Components.interfaces.calIPrintFormatter)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return this;
};

FGcalYearPrinter.prototype.getName =
function FGyearPrint_getName() {
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                    .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://fgprinters/locale/fgprinters.properties");
    var retName = "FG " + props.GetStringFromName("yearview");
    return retName;
};
FGcalYearPrinter.prototype.__defineGetter__("name", FGcalYearPrinter.prototype.getName);

FGcalYearPrinter.prototype.formatToHtml =
function FGyearPrint_format(aStream, aStart, aEnd, aCount, aItems, aTitle) {

    var html = <html/>
    html.appendChild(
            <head>
                <title>{aTitle}</title>
                <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>
                <style type='text/css'/>
            </head>);
    var fontsizetitle = getPrefSafe("fgprinters.year.fontsizetitle", 26);
    var fontsizeday = getPrefSafe("fgprinters.year.fontsizemonth", 16);
    var dnbk = getPrefSafe("fgprinters.headcolour", "#f0f0f0");
    var oomc = getPrefSafe("fgprinters.othermonthcolour", "#EEEEFF");
    var wec = getPrefSafe("fgprinters.weekendcolour", "#FFFFEE");
    var sansserif = getPrefSafe("fgprinters.sansserif", true);
    var printtitle = getPrefSafe("fgprinters.printtitle", false);
    var fsizeglobaltitle = getPrefSafe("fgprinters.fontsizetitle", 26);
    var fsizesubtitle = getPrefSafe("fgprinters.fontsizesubtitle", 20);
    var subTitle = getPrefSafe("fgprinters.subtitle", "");
    var halfYear = getPrefSafe("fgprinters.year.halfyear", false);
    var horView = getPrefSafe("fgprinters.year.horizontal", false);
    var sepDayNames = getPrefSafe("fgprinters.year.seperatedaynames", true);
    var fsize = getPrefSafe("fgprinters.year.fontsize", 10);
    var firstLetter = getPrefSafe("fgprinters.year.firstletter", false);
    var printWeekDays = getPrefSafe("fgprinters.year.weekdays", false);

    var dnStyle = 'font-size: ' + fsize + 'px;';

    html.head.style = ".main-table { font-size: "+fontsizetitle+"px; font-weight: bold; }\n";
    html.head.style += ".main-title { font-size: "+fsizeglobaltitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".sub-title { font-size: "+fsizesubtitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".day-box { border: 1px solid black; vertical-align: top; }\n";
    html.head.style += ".day-box-sun { border-top: 1px solid black; vertical-align: top; }\n";
    html.head.style += ".day-off { background-color: "+wec+" !important; }\n";
    if (sansserif) { html.head.style += "body, table, tr, td, p { font-family: Arial, sans-serif; }\n"; }

    // If aStart or aEnd weren't passed in, we need to calculate them based on
    // aItems data.

    var start = aStart;
    var end = aEnd;
    if (!start || !end) {
        for each (var item in aItems) {
            var itemStart = item.startDate || item.entryDate;
            var itemEnd = item.endDate || item.dueDate;
            if (!start || (itemStart && start.compare(itemStart) == 1)) {
                start = itemStart;
            }
            if (!end || (itemEnd && end.compare(itemEnd) == -1)) {
                end = itemEnd;
            }
        }
    }
    // Play around with aStart and aEnd to determine the minimal number of
    // months we can show to still technically meet their requirements.  This
    // is most useful when someone printed 'Current View' in the month view. If
    // we take the aStart and aEnd literally, we'll print 3 months (because of
    // the extra days at the start/end), but we should avoid that.
    //
    // Basically, we check whether aStart falls in the same week as the start
    // of a month (ie aStart  is Jan 29, which often is in the same week as
    // Feb 1), and similarly whether aEnd falls in the same week as the end of
    // a month.
    var weekStart = getPrefSafe("calendar.week.start", 0);
    maybeNewStart = start.clone();
    maybeNewStart.day = 1;
    maybeNewStart.month = start.month+1;
    try {
        maybeNewStart.normalize();
    } catch(ex) {
    }

    var date = start.clone();
    date.month = 0;
    try {
        date.normalize();
    } catch(ex) {
    }

    var realEnd = end.clone();
    realEnd.month = 12;
    realEnd.day = 10;
    try {
        realEnd.normalize();
    } catch(ex) {
    }

    var body = <body/>

    if (printtitle && aTitle) {
        body.appendChild(<p class="main-title">{aTitle}</p>);
    }

    if (subTitle) {
        body.appendChild(<p class="sub-title">{subTitle}</p>);
    }

    var year = <table width='100%'/>;
    var allMonths = <tr style='page-break-inside: avoid;' />;
    var first = 2;

    while (date.compare(realEnd) < 0) {
        if (first) {
            if (first == 1) {
                year.appendChild(allMonths);
                allMonths = <tr style='page-break-inside: avoid;' />;
            }
            if (first == 3) {
                year.appendChild(allMonths);
                allMonths = <tr style='page-break-inside: avoid;' />;
            }
            if ((first == 1) || (first == 2)) {
                var head = <tr/>;
                var fsize = getPrefSafe("fgprinters.year.fontsizetitle", 26);
                var tStyle = 'font-size: ' + fsize + 'px;';
                head.appendChild(<td colspan="38" align="center" style={tStyle}>{date.year}</td>);
                year.appendChild(head);
                if (horView && sepDayNames && printWeekDays) {
                    var daysTr = <tr style='page-break-inside: avoid;' />;
                    daysTr.appendChild(<td style='color: transparent;'>A </td>);
                    for (var i = 0; i < 37; i++) {
                        var dayName = calGetString("dateFormat", "day."+ (((i + weekStart) % 7)+1) + ".Mmm");
                        if (firstLetter) {
                            dayName = dayName[0];
                        }
                        daysTr.appendChild(<td style={dnStyle}>{dayName}</td>);
                    }
                    year.appendChild(daysTr);
                }
            }
            first = 0;
        }

        if (horView) {
            allMonths = this.createMonthTable(date);
        } else {
            if (halfYear) {
                var thisMonth = <td valign="top" width="16%"/>;
            } else {
                var thisMonth = <td valign="top" width="8%"/>;
            }
            thisMonth.appendChild(this.createMonthTable(date));
            allMonths.appendChild(thisMonth);
        }
        date.month++;
        try {
            date.normalize();
        } catch(ex) {
        }
        if (date.month == 0) {
            first = 1;
        }
        if ((date.month == 6) && halfYear) {
            first = 3;
        }
        if (horView) {
            first = 3;
        }
    }
    year.appendChild(allMonths);
    body.appendChild(year);
    html.appendChild(body);
    
    var convStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                               .getService(Components.interfaces.nsIConverterOutputStream);
    convStream.init(aStream, 'UTF-8', 0, 0x0000);
    convStream.writeString(html.toXMLString());
//    Components.utils.reportError(html.toXMLString());
};


function getContrastingTextColor(bgColor)
{
    var calcColor = bgColor.replace(/#/g, "");
    var red = parseInt(calcColor.substring(0, 2), 16);
    var green = parseInt(calcColor.substring(2, 4), 16);
    var blue = parseInt(calcColor.substring(4, 6), 16);

    // Calculate the L(ightness) value of the HSL color system.
    // L = (max(R, G, B) + min(R, G, B)) / 2
    var max = Math.max(Math.max(red, green), blue);
    var min = Math.min(Math.min(red, green), blue);
    var lightness = (max + min) / 2;

    // Consider all colors with less than 50% Lightness as dark colors
    // and use white as the foreground color; otherwise use black.
    // Actually we use a threshold a bit below 50%, so colors like
    // #FF0000, #00FF00 and #0000FF still get black text which looked
    // better when we tested this.
    if (lightness < 120) {
        return "white";
    }
    
    return "black";
}

FGcalYearPrinter.prototype.createMonthTable =
function FGyearPrint_doOtherMonth(aMonth) {
    var weekStart = getPrefSafe("calendar.week.start", 0);

    var monthName = calGetString("dateFormat", "month." + (aMonth.month +1)+ ".name"); // + " " + aMonth.year;

    var fsize = getPrefSafe("fgprinters.year.fontsize", 10);
    var tStyle = 'font-size: ' + fsize + 'px;';
    var fsizem = getPrefSafe("fgprinters.year.fontsizemonth", 16);
    var mStyle = 'font-size: ' + fsizem + 'px;';
    var firstLetter = getPrefSafe("fgprinters.year.firstletter", false);
    var printWeekDays = getPrefSafe("fgprinters.year.weekdays", false);
    var alignDays = getPrefSafe("fgprinters.year.aligndays", false);
    var horView = getPrefSafe("fgprinters.year.horizontal", false);
    var sepDayNames = getPrefSafe("fgprinters.year.seperatedaynames", true);

    var monthTable;

    if (horView) {
        monthTable = <tr style={tStyle}><td align='center' width='0%' style={mStyle}>{monthName}</td></tr>;
    } else {
        monthTable = <table width='100%' rules='both' style={tStyle}><tr><td align='center' style={mStyle}>{monthName}</td></tr></table>;
    }

    var firstDate = aMonth.startOfMonth.clone();

    var lastDate = aMonth.endOfMonth.clone();
    firstDate.isDate = true;
    lastDate.isDate = true;

    var nameStyle = "width: "+fsize*3+"px; float: left; text-align: left;";

    var date = firstDate.clone();
    var itemListIndex = 0;
    var weekRow = <tr/>;
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];
    var daysOff = new Array();
    for (var i in prefNames) {
        if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
            daysOff.push(Number(i));
        }
    }

    if (horView) {
        if ((alignDays) || ((sepDayNames) && (printWeekDays))) {
            var startWeekday = date.weekday;
            if (startWeekday < weekStart) {
                startWeekday += 7;
            }
            for (var i = weekStart; i < startWeekday; i++) {
                var fakeday = <td width='3%' style='border: 1px solid transparent; vertical-align: center; color: transparent;'>A </td>;
                monthTable.appendChild(fakeday);
            }
        }
        while (date.compare(lastDate) != 1) {
            var myClass = 'day-box';
            if (daysOff.some(function(a) { return a == date.weekday; })) {
                myClass += ' day-off';
            }
            if ((printWeekDays) && !(sepDayNames)) {
                var dayName = calGetString("dateFormat", "day."+ (date.weekday + 1) + ".Mmm");
                if (firstLetter) {
                    dayName = dayName[0];
                }
                var day = <td width='3%' class={myClass} style='text-align: right; vertical-align: center;'><div style={nameStyle}>{dayName}</div> {date.day}</td>;
            } else {
                var day = <td width='3%' class={myClass} style='text-align: left; vertical-align: center;'>{date.day}</td>;
            }
            monthTable.appendChild(day);
            date.day++;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
    } else {
        var monthCol = <td/>;
        if (alignDays) {
            var startWeekday = date.weekday;
            if (startWeekday < weekStart) {
                startWeekday += 7;
            }
            for (var i = weekStart; i < startWeekday; i++) {
                var fakeday = <div style='border: 1px solid transparent; vertical-align: top; color: transparent;'>A </div>;
                monthCol.appendChild(fakeday);
            }
        }
        while (date.compare(lastDate) != 1) {
            var myClass = 'day-box';
            if (daysOff.some(function(a) { return a == date.weekday; })) {
                myClass += ' day-off';
            }
            if (printWeekDays) {
                var dayName = calGetString("dateFormat", "day."+ (date.weekday + 1) + ".Mmm");
                if (firstLetter) {
                    dayName = dayName[0];
                }
                var day = <div class={myClass} style='text-align: right;'><div style={nameStyle}>{dayName}</div> {date.day}</div>;
            } else {
                var day = <div class={myClass} style='text-align: left;'>{date.day}</div>;
            }
            monthCol.appendChild(day);
            date.day++;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
        weekRow.appendChild(monthCol);
        monthTable.appendChild(weekRow);
    }
    return monthTable;
};



/**
 * Normal get*Pref calls will throw if the pref is undefined.  This function
 * will get a bool, int, or string pref.  If the pref is undefined, it will
 * return aDefault.
 *
 * @param aPrefName   the (full) name of preference to get
 * @param aDefault    (optional) the value to return if the pref is undefined
 */

function getPrefSafe(aPrefName, aDefault) {
    const nsIPrefBranch = Components.interfaces.nsIPrefBranch;
    const prefB = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(nsIPrefBranch);
    switch (prefB.getPrefType(aPrefName)) {
        case nsIPrefBranch.PREF_BOOL:
            return prefB.getBoolPref(aPrefName);
        case nsIPrefBranch.PREF_INT:
            return prefB.getIntPref(aPrefName);
        case nsIPrefBranch.PREF_STRING:
            return prefB.getCharPref(aPrefName);
        default: // includes nsIPrefBranch.PREF_INVALID
            return aDefault;
    }
}


/**
 * Gets the value of a string in a .properties file
 *
 * @param aBundleName  the name of the properties file.  It is assumed that the
 *                     file lives in chrome://calendar/locale/
 * @param aStringName the name of the string within the properties file
 */
function calGetString(aBundleName, aStringName) {
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                        .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://calendar/locale/"+aBundleName+".properties");
    return props.GetStringFromName(aStringName);
}

function log(what) {
    var consoleSvc = Components.classes["@mozilla.org/consoleservice;1"].
                     getService(Components.interfaces.nsIConsoleService);
    consoleSvc.logStringMessage(what);    
}

/**
 * Function to get the (cached) best guess at a user's default timezone.  We'll
 * use the value of the calendar.timezone.local preference, if it exists.  If
 * not, we'll do our best guess.
 *
 * @return user's default timezone.
 */
function calendarDefaultTimezone() {
    if (calendarDefaultTimezone.mTz === undefined) {
        var prefTzid = getPrefSafe("calendar.timezone.local", null);
        var tzid = prefTzid;
        if (!tzid) {
            tzid = guessSystemTimezone();
        }
        calendarDefaultTimezone.mTz = getTimezoneService().getTimezone(tzid);
        ASSERT(calendarDefaultTimezone.mTz, "timezone not found: " + tzid);
        // Update prefs if necessary:
        if (calendarDefaultTimezone.mTz && calendarDefaultTimezone.mTz.tzid != prefTzid) {
            setPref("calendar.timezone.local", "CHAR", calendarDefaultTimezone.mTz.tzid);
        }
    }
    return calendarDefaultTimezone.mTz;
}

function getTimezoneService() {
    if (getTimezoneService.mObject === undefined) {
        getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"]
                                               .getService(Components.interfaces.calITimezoneService);
    }
    return getTimezoneService.mObject;
}
